@@ -76,28 +76,30 @@ static LPCWSTR  STYPE_INTERACTIVE = L"interactive";
 static LPWSTR       _service_name = NULL;
 /* Allowed procrun commands */
 static LPCWSTR _commands[] = {
-    L"TS",      /* 1 Run Service as console application (default)*/
-    L"RS",      /* 2 Run Service */
-    L"ES",      /* 3 Execute start */
-    L"SS",      /* 4 Stop Service */
-    L"US",      /* 5 Update Service parameters */
-    L"IS",      /* 6 Install Service */
-    L"DS",      /* 7 Delete Service */
-    L"?",       /* 8 Help */
-    L"VS",      /* 9 Version */
+    L"TS",      /*  1 Run Service as console application (default)*/
+    L"RS",      /*  2 Run Service */
+    L"ES",      /*  3 Execute start */
+    L"SS",      /*  4 Stop Service */
+    L"US",      /*  5 Update Service parameters */
+    L"IS",      /*  6 Install Service */
+    L"DS",      /*  7 Delete Service */
+    L"PS",      /*  8 Print Service Configuration */
+    L"?",       /*  9 Help */
+    L"VS",      /* 10 Version */
     NULL
 };
 
 static LPCWSTR _altcmds[] = {
-    L"run",         /* 1 Run Service as console application (default)*/
-    L"service",     /* 2 Run Service */
-    L"start",       /* 3 Start Service */
-    L"stop",        /* 4 Stop Service */
-    L"update",      /* 5 Update Service parameters */
-    L"install",     /* 6 Install Service */
-    L"delete",      /* 7 Delete Service */
-    L"help",        /* 8 Help */
-    L"version",     /* 9 Version */
+    L"run",         /*  1 Run Service as console application (default)*/
+    L"service",     /*  2 Run Service */
+    L"start",       /*  3 Start Service */
+    L"stop",        /*  4 Stop Service */
+    L"update",      /*  5 Update Service parameters */
+    L"install",     /*  6 Install Service */
+    L"delete",      /*  7 Delete Service */
+	L"print",       /*  8 Print Service Configuration */
+    L"help",        /*  9 Help */
+    L"version",     /* 10 Version */
     NULL
 };
 
@@ -282,7 +284,7 @@ DWORD WINAPI eventThread(LPVOID lpParam)
     }
     ExitThread(0);
     return 0;
-	UNREFERENCED_PARAMETER(lpParam);
+    UNREFERENCED_PARAMETER(lpParam);
 }
 
 /* redirect console stdout/stderr to files
@@ -324,13 +326,13 @@ static BOOL redirectStdStreams(APX_STDWRAP *lpWrapper, LPAPXCMDLINE lpCmdline)
             DeleteFileW(lpWrapper->szStdOutFilename);
         if ((lpWrapper->fpStdOutFile = _wfsopen(lpWrapper->szStdOutFilename,
                                                L"a",
-											   _SH_DENYNO))) {
+                                               _SH_DENYNO))) {
             _dup2(_fileno(lpWrapper->fpStdOutFile), 1);
             *stdout = *lpWrapper->fpStdOutFile;
             setvbuf(stdout, NULL, _IONBF, 0);
         }
         else {
-        	lpWrapper->szStdOutFilename = NULL;
+            lpWrapper->szStdOutFilename = NULL;
         }
     }
     if (lpWrapper->szStdErrFilename) {
@@ -350,7 +352,7 @@ static BOOL redirectStdStreams(APX_STDWRAP *lpWrapper, LPAPXCMDLINE lpCmdline)
             DeleteFileW(lpWrapper->szStdErrFilename);
         if ((lpWrapper->fpStdErrFile = _wfsopen(lpWrapper->szStdErrFilename,
                                                L"a",
-											   _SH_DENYNO))) {
+                                               _SH_DENYNO))) {
             _dup2(_fileno(lpWrapper->fpStdErrFile), 2);
             *stderr = *lpWrapper->fpStdErrFile;
             setvbuf(stderr, NULL, _IONBF, 0);
@@ -382,6 +384,7 @@ static void printUsage(LPAPXCMDLINE lpCmdline, BOOL isHelp)
     fwprintf(stderr, L"  start   [ServiceName]  Start Service\n");
     fwprintf(stderr, L"  stop    [ServiceName]  Stop Service\n");
     fwprintf(stderr, L"  run     [ServiceName]  Run Service as console application\n");
+    fwprintf(stderr, L"  print   [ServiceName]  Display the command to (re-)create the current configuration\n");
     fwprintf(stderr, L"  pause   [Num Seconds]  Sleep for n Seconds (defaults to 60)\n");
     fwprintf(stderr, L"  version                Display version\n");
     fwprintf(stderr, L"  Options:\n");
@@ -395,12 +398,12 @@ static void printVersion(void)
 {
     fwprintf(stderr, L"Apache Commons Daemon Service Runner version %S/Win%d (%S)\n",
             PRG_VERSION, PRG_BITS, __DATE__);
-    fwprintf(stderr, L"Copyright (c) 2000-2019 The Apache Software Foundation.\n\n"
+    fwprintf(stderr, L"Copyright (c) 2000-2021 The Apache Software Foundation.\n\n"
                      L"For bug reporting instructions, please see:\n"
                      L"<URL:https://issues.apache.org/jira/browse/DAEMON>.");
 }
 
-/* Display configuration parameters */
+/* Display comamnd line parameters */
 static void dumpCmdline()
 {
     int i = 0;
@@ -430,14 +433,14 @@ void apxSetInprocEnvironment()
 
     hmodUcrt = LoadLibraryExA("ucrtbase.dll", NULL, LOAD_LIBRARY_SEARCH_SYSTEM32);
     if (hmodUcrt != NULL) {
-    	wputenv_ucrt =  (WPUTENV) GetProcAddress(hmodUcrt, "_wputenv");
+        wputenv_ucrt =  (WPUTENV) GetProcAddress(hmodUcrt, "_wputenv");
     }
 
     for (p = SO_ENVIRONMENT; *p; p++) {
         e = apxExpandStrW(gPool, p);
         _wputenv(e);
         if (wputenv_ucrt != NULL) {
-        	wputenv_ucrt(e);
+            wputenv_ucrt(e);
         }
         apxFree(e);
         while (*p)
@@ -481,6 +484,64 @@ static void setInprocEnvironment9(LPCWSTR szOptions9)
     apxFree(e);
 }
 
+/* Sets environment variables required by some Java options
+ * Currently only Native Memory Tracking
+ */
+static void setInprocEnvironmentOptions(LPCWSTR szOptions)
+{
+    LPCWSTR p;
+    DWORD len;
+    LPWSTR e;
+
+    apxLogWrite(APXLOG_MARK_DEBUG "Checking Java options for environment variable requirements");
+
+    p = szOptions;
+    while (*p) {
+        apxLogWrite(APXLOG_MARK_DEBUG "Checking '%S' for environment variable requirements", p);
+        if (wcsncmp(p, L"-XX:NativeMemoryTracking=", 25) == 0) {
+            apxLogWrite(APXLOG_MARK_DEBUG "Match found '%S'", p);
+            /* Advance 25 characters to the start of the value */
+            p += 25;
+            apxLogWrite(APXLOG_MARK_DEBUG "Setting is '%S'", p);
+            /* Ignore setting if it is off */
+            if (wcsncmp(p, L"off", 3)) {
+                apxLogWrite(APXLOG_MARK_DEBUG "Creating environment entry");
+                /* Allocated space for the setting value */
+                len = lstrlen(p);
+                /* Expand space to include env var name less pid and '=' */
+                len += 11;
+                /* Expand spave to include pid (4 bytes, signed - up to 10 characters */
+                len += 10;
+                /* Expand space to include the null terminator */
+                len++;
+
+                /* Allocate the space */
+                e = apxPoolCalloc(gPool, len * sizeof(WCHAR));
+
+                /* Create the environment variable needed by NMT */
+                swprintf_s(e, len, L"NMT_LEVEL_%d=%s", GetCurrentProcessId(), p);
+
+                apxLogWrite(APXLOG_MARK_DEBUG "Created environment entry '%S'", e);
+                /* Set the environment variable */
+               _wputenv(e);
+
+               apxFree(e);
+            }
+            return;
+        }
+
+        /* advance to the terminating null */
+        while(*p) {
+          p++;
+        }
+
+        /* advance to the start of the next entry
+         * will be null if there are no more entries
+         */
+        p++;
+    }
+}
+
 /* Load the configuration from Registry
  * loads only nonspecified items
  */
@@ -598,16 +659,64 @@ static BOOL saveConfiguration(LPAPXCMDLINE lpCmdline)
                                _options[i].dwValue);
         else if (_options[i].dwType & APXCMDOPT_MSZ)
             apxRegistrySetMzStrW(hRegistry, APXREG_PARAMSOFTWARE,
-                                 _options[i].szSubkey,
-                                 _options[i].szRegistry,
-                                 _options[i].szValue,
-                                 _options[i].dwValue);
+                               _options[i].szSubkey,
+                               _options[i].szRegistry,
+                               _options[i].szValue,
+                               _options[i].dwValue);
         ++i;
     }
     apxCloseHandle(hRegistry);
     return TRUE;
 }
 
+/* Display current configuration */
+static BOOL printConfig(LPAPXCMDLINE lpCmdline)
+{
+    int i = 0;
+
+    if (!loadConfiguration(lpCmdline)) {
+		return FALSE;
+	}
+
+    fwprintf(stderr, L"%s.exe update ",lpCmdline->szExecutable);
+    while (_options[i].szName) {
+		if (_options[i].dwType & APXCMDOPT_INT) {
+			fwprintf(stderr, L"--%s %d ", _options[i].szName, _options[i].dwValue);
+		} else if (_options[i].dwType & APXCMDOPT_MSZ) {
+			if (_options[i].szValue) {
+				BOOL first = TRUE;
+				LPCWSTR p = _options[i].szValue;
+    	        fwprintf(stderr, L"--%s \"", _options[i].szName);
+
+    	        while (*p) {
+    	        	if (first) {
+    	        		first = FALSE;
+    	        	} else {
+    	        		fwprintf(stderr, L"#");
+    	        	}
+    	        	// Skip to terminating NULL for this value
+    	        	while (p[0]) {
+    	        		if (p[0] == L'#') {
+    	        			fwprintf(stderr, L"'%c'", p[0]);
+    	        		} else {
+    	        			fwprintf(stderr, L"%c", p[0]);
+    	        		}
+    	        		++p;
+    	        	}
+    	        	// Move to start of next value
+    	        	++p;
+    	        }
+    	        fwprintf(stderr, L"\" ");
+			}
+        } else if (_options[i].szValue) {
+            fwprintf(stderr, L"--%s \"%s\" ", _options[i].szName, _options[i].szValue);
+        }
+        // NULL options are ignored
+        ++i;
+    }
+    return TRUE;
+}
+
 /* Operations */
 static BOOL docmdInstallService(LPAPXCMDLINE lpCmdline)
 {
@@ -630,8 +739,21 @@ static BOOL docmdInstallService(LPAPXCMDLINE lpCmdline)
         dwStart = SERVICE_AUTO_START;
     /* Check the service type */
     if ((ST_TYPE & APXCMDOPT_FOUND) &&
-        lstrcmpiW(SO_TYPE, STYPE_INTERACTIVE) == 0)
-        dwType |= SERVICE_INTERACTIVE_PROCESS;
+        lstrcmpiW(SO_TYPE, STYPE_INTERACTIVE) == 0) {
+
+        // Need to run as LocalSystem to set the interactive flag
+        LPCWSTR su = NULL;
+        if (ST_SUSER & APXCMDOPT_FOUND) {
+            su = SO_SUSER;
+        }
+        if (su && lstrcmpiW(su, L"LocalSystem") == 0) {
+            dwType |= SERVICE_INTERACTIVE_PROCESS;
+        } else {
+            apxLogWrite(APXLOG_MARK_ERROR
+                    "The parameter '--Type interactive' is only valid with '--ServiceUser LocalSystem'");
+            return FALSE;
+        }
+    }
 
     /* Check if --Install is provided */
     if (!IS_VALID_STRING(SO_INSTALL)) {
@@ -879,7 +1001,7 @@ static BOOL docmdUpdateService(LPAPXCMDLINE lpCmdline)
         rv = (rv && apxServiceSetOptions(hService,
                                          dwType,
                                          dwStart,
-										 bDelayedStart,
+                                         bDelayedStart,
                                          SERVICE_NO_CHANGE));
 
         apxLogWrite(APXLOG_MARK_INFO "Updated service '%S'.",
@@ -968,7 +1090,7 @@ BOOL child_callback(APXHANDLE hObject, UINT uMsg,
             fputc(ch, stdout);
     }
     return TRUE;
-	UNREFERENCED_PARAMETER(hObject);
+    UNREFERENCED_PARAMETER(hObject);
 }
 
 static int onExitStop(void)
@@ -1051,11 +1173,11 @@ static DWORD WINAPI serviceStop(LPVOID lpParameter)
         else {
             if (lstrcmpA(_jni_sclass, "java/lang/System") == 0) {
                 reportServiceStatus(SERVICE_STOP_PENDING, NO_ERROR, 20 * 1000);
-                apxLogWrite(APXLOG_MARK_DEBUG "Forcing Java JNI System.exit worker to finish...");
+                apxLogWrite(APXLOG_MARK_DEBUG "Forcing Java JNI System.exit() worker to finish...");
                 return 0;
             }
             else {
-                apxLogWrite(APXLOG_MARK_DEBUG "Waiting for Java JNI stop worker to finish...");
+                apxLogWrite(APXLOG_MARK_DEBUG "Waiting for Java JNI stop worker to finish for %s:%s...", _jni_sclass, _jni_smethod);
                 apxJavaWait(hWorker, INFINITE, FALSE);
                 apxLogWrite(APXLOG_MARK_DEBUG "Java JNI stop worker finished.");
             }
@@ -1230,6 +1352,12 @@ static DWORD serviceStart()
             /* Add LibraryPath to the PATH */
            apxAddToPathW(gPool, SO_LIBPATH);
         }
+        /* Some options require additional environment settings to be in place
+         * before Java is started
+         */
+        if (IS_VALID_STRING(SO_JVMOPTIONS)) {
+        	setInprocEnvironmentOptions(SO_JVMOPTIONS);
+        }
         /* Create the JVM global worker */
         gWorker = apxCreateJava(gPool, _jni_jvmpath, SO_JAVAHOME);
         if (IS_INVALID_HANDLE(gWorker)) {
@@ -1274,6 +1402,7 @@ static DWORD serviceStart()
         }
         /* Set the environment using putenv, so JVM can use it */
         apxSetInprocEnvironment();
+        /* Java 9 specific options need to be set via an environment variable */
         setInprocEnvironment9(SO_JVMOPTIONS9);
         /* Redirect process */
         gWorker = apxCreateProcessW(gPool,
@@ -1386,7 +1515,7 @@ void WINAPI service_ctrl_handler(DWORD dwCtrlCode)
             /* Stop the service asynchronously */
             stopThread = CreateThread(NULL, 0,
                                       serviceStop,
-                                      (LPVOID)dwCtrlCode,
+                                      (LPVOID)SERVICE_CONTROL_STOP,
                                       0, &threadId);
             CloseHandle(stopThread);
             return;
@@ -1645,8 +1774,8 @@ cleanup:
     reportServiceStatusStopped(rc);
     gExitval = rc;
     return;
-	UNREFERENCED_PARAMETER(argc);
-	UNREFERENCED_PARAMETER(argv);
+    UNREFERENCED_PARAMETER(argc);
+    UNREFERENCED_PARAMETER(argv);
 }
 
 
@@ -1660,7 +1789,7 @@ BOOL docmdDebugService(LPAPXCMDLINE lpCmdline)
     apxLogWrite(APXLOG_MARK_INFO "Debug service finished with exit code %d.", gExitval);
     SAFE_CLOSE_HANDLE(gPidfileHandle);
     if (gPidfileName) {
-   	    DeleteFileW(gPidfileName);
+        DeleteFileW(gPidfileName);
     }
     return gExitval == 0 ? TRUE : FALSE;
 }
@@ -1686,7 +1815,7 @@ BOOL docmdRunService(LPAPXCMDLINE lpCmdline)
     }
     SAFE_CLOSE_HANDLE(gPidfileHandle);
     if (gPidfileName) {
-   	    DeleteFileW(gPidfileName);
+        DeleteFileW(gPidfileName);
     }
     return rv;
 }
@@ -1817,10 +1946,13 @@ void __cdecl main(int argc, char **argv)
             if (!docmdDeleteService(lpCmdline))
                 rv = 9;
         break;
-        case 8: /* Print Usage and exit */
+        case 8: /* Print Configuration and exit */
+            printConfig(lpCmdline);
+        break;
+        case 9: /* Print Usage and exit */
             printUsage(lpCmdline, TRUE);
         break;
-        case 9: /* Print version and exit */
+        case 10: /* Print version and exit */
             printVersion();
         break;
         default:
